"""Info solver
==============

Container to encode the solver class hierarchy.

.. autofunction:: create_info_simul

.. autoclass:: InfoSolverCore
   :members:
   :private-members:

"""

from fluiddyn.util import import_class
from fluiddyn.util.paramcontainer import ParamContainer


def create_info_simul(info_solver, params):
    """Create a ParamContainer instance gathering info_solver and params."""
    info = ParamContainer(tag="info_simul")
    info._set_as_child(info_solver)
    info._set_as_child(params)
    return info


class InfoSolverCore(ParamContainer):
    """Contain the information on a solver.

    Parameters
    ----------

    only_root: bool (False)

      If True, only initialize the root level.

    Attributes
    ----------

    _modificators: list

      List of functions modifying the instance at the end of ``__init__``.

    _extenders: list

      List of extenders (see :mod:`fluidsim_core.extend_simul`).

    """

    def __init__(self, only_root=False, **kargs):
        if len(kargs) == 0 or ("path_file" in kargs and "tag" not in kargs):
            kargs["tag"] = "solver"

        super().__init__(**kargs)

        if (
            "tag" in kargs
            and kargs["tag"] == "solver"
            and "path_file" not in kargs
        ):
            self._init_root()

            if not only_root:
                self.complete_with_classes()

                if hasattr(self, "_modificators"):
                    for modif_info_solver in self._modificators:
                        modif_info_solver(self)

                    self._set_attrib(
                        "extenders",
                        [
                            f"{extender._module_name}.{extender.__name__}"
                            for extender in self._extenders
                        ],
                    )

    def _init_root(self):
        self._set_attribs(
            {
                "module_name": "fluidsim_core.solver",
                "class_name": "SimulCore",
                "short_name": "Core",
            }
        )

        _ = self._set_child("classes")

    def _set_attrib(self, key, value):
        """Add an attribute to the container."""
        if isinstance(value, dict) and all(
            k in value for k in ("module_name", "class_name")
        ):
            from warnings import warn

            warn(
                f"A class {key} was specified in the InfoSolver "
                "instance using _set_attrib method. Perhaps you meant \n\n"
                f"_set_child('{key}', {value}) \n\ninstead?"
            )

        super()._set_attrib(key, value)

    def import_classes(self):
        """Import the classes and return a dictionary."""
        if getattr(self, "_cached_imported_classes", None) is not None:
            return self._cached_imported_classes

        dict_classes = {}
        tags = self.classes._tag_children
        if len(tags) == 0:
            self._set_internal_attr("_cached_imported_classes", dict_classes)
            return dict_classes

        for tag in tags:
            cls = self.classes.__dict__[tag]
            try:
                module_name = cls.module_name
                class_name = cls.class_name
            except AttributeError:
                pass
            else:
                Class = import_class(module_name, class_name)
                dict_classes[cls._tag] = Class

        self._set_internal_attr("_cached_imported_classes", dict_classes)
        return dict_classes

    def complete_with_classes(self):
        """Populate info solver by executing ``_complete_info_solver`` class
        methods
        """
        dict_classes = self.import_classes()
        for Class in list(dict_classes.values()):
            if hasattr(Class, "_complete_info_solver"):
                Class._complete_info_solver(self)



from fluidsim_core.loader import import_cls_simul
from fluidsim_core.tests.solver import SimulTest


def test_load_cls_simul():
    Simul = import_cls_simul("test", "fluidsim_core.tests")
    assert Simul is SimulTest



"""The Simulation class API.

.. autoclass:: SimulCore
   :members: _complete_params_with_default, create_default_params, __init__

"""

from abc import ABC, abstractclassmethod, abstractmethod, abstractstaticmethod

from .info import InfoSolverCore, create_info_simul
from .params import Parameters


__all__ = ["SimulCore"]


class SimulCore(ABC):
    """Represent a solver. This is an abstract base class describing the
    bare minimum number of methods to have a working solver.

    """

    InfoSolver = InfoSolverCore
    Parameters = Parameters

    @abstractstaticmethod
    def _complete_params_with_default(params):
        """A static method used to complete the *params* container."""
        attribs = {}
        params._set_attribs(attribs)
        #  params._set_doc("""Describe docstrings for params here.""")

    @classmethod
    def create_default_params(cls):
        """Sets an ``info_solver`` instance as a class attribute and returns a
        *params* container populated with default values.

        """
        cls.info_solver = cls.InfoSolver()
        return cls.Parameters._create_params(cls.info_solver)

    @abstractmethod
    def __init__(self, params):
        """Instantiate Simulation class and optionally all classes as specified
        by :any:`fluidsim_core.info.InfoSolverCore.import_classes`"""
        if not hasattr(self, "info_solver") or not isinstance(
            self.info_solver, self.InfoSolver
        ):
            # The condition `not isinstance(self.info_solver, self.InfoSolver)`
            # is needed in some rare situations (mostly testing) for which
            # info_solver comes from an inherited class.
            self.info_solver = self.InfoSolver()

        if not isinstance(params, self.Parameters):
            raise TypeError(
                f"params should be a {self.Parameters} instance, "
                f"not {type(params)}"
            )

        self.params = params
        self.info = create_info_simul(self.info_solver, params)

        # Instantiate other classes as follows
        # ------------------------------------
        # dict_classes = self.info_solver.import_classes()
        # Operators = dict_classes["Operators"]
        # self.oper = Operators(params=params)
        # ..



"""For command fluidsim-ipy-load

.. autofunction:: start_ipython_load_sim

"""

import argparse
from textwrap import dedent


def start_ipython_load_sim(load_import="from fluidsim import load"):
    """Start IPython and load a simulation"""

    parser = argparse.ArgumentParser(
        prog="ipy-load",
        description="Start IPython and load a simulation",
    )
    parser.add_argument("path_dir", nargs="?", default=None)
    args = parser.parse_args()

    from IPython import start_ipython

    argv = ["--matplotlib", "-i", "-c"]

    code = dedent(
        f"""
        import numpy as np
        import matplotlib.pyplot as plt
        import pandas as pd
        {load_import}
        print("Loading simulation")
        sim = load()
        params = sim.params
        print("`sim`, `params`, `np`, `plt` and `pd` variables are available")
    """
    )
    lines = code.strip().split("\n")

    if args.path_dir is not None:
        from fluidsim_core.paths import find_path_result_dir

        path_dir = find_path_result_dir(args.path_dir)
        lines.insert(0, f"import os; os.chdir('{path_dir}')")

    argv.append("; ".join(lines))
    start_ipython(argv=argv)



"""Mechanism to extend a Simul class with just a simple class
=============================================================

.. autofunction:: extend_simul_class

.. autoclass:: SimulExtender
   :members:
   :private-members:

"""

from abc import ABCMeta, abstractclassmethod
from logging import warn
from pathlib import Path

import h5py

from fluiddyn.util.paramcontainer import ParamContainer
from fluiddyn.util import import_class
from fluiddyn.util import mpi


def extend_simul_class(Simul, extenders):
    """Extend a Simul class with "Simul extenders".

    "Simul extenders" are classes deriving from
    :class:`fluidsim.extend_simul.SimulExtender`.

    """

    if isinstance(extenders, type):
        extenders = [extenders]

    class NewInfoSolver(Simul.InfoSolver):
        pass

    if not hasattr(NewInfoSolver, "_modificators"):
        NewInfoSolver._modificators = []
        NewInfoSolver._extenders = []

    for extender in extenders:
        modif_info_solver = extender.get_modif_info_solver()
        NewInfoSolver._modificators.append(modif_info_solver)
        NewInfoSolver._extenders.append(extender)

    class NewSimul(Simul):
        InfoSolver = NewInfoSolver

    return NewSimul


class SimulExtender(metaclass=ABCMeta):
    """Abstract class to define a "Simul extender"

    Simul extenders are classes that can extend a ``Simul`` class to change its
    behavior for some simulations.

    This class is meant to be subclassed. The child class has to contain one
    class attribute ``_module_name`` and two class methods
    ``get_modif_info_solver`` and ``complete_params_with_default``. An example
    can be found in the module
    :mod:`fluidsim.extend_simul.spatial_means_regions_milestone`.

    """

    @abstractclassmethod
    def get_modif_info_solver(cls):
        """Create a function to modify ``info_solver``.

        Note that this function is called when the object ``info_solver`` has
        not yet been created (and cannot yet be modified)! This is why one
        needs to create a function that will be called later to modify
        ``info_solver``.

        """

    @abstractclassmethod
    def complete_params_with_default(cls, params):
        """Should complete the simul parameters"""

    @classmethod
    def _complete_params_with_default(cls, params):
        # so that writers of SimulExtender classes can write
        # complete_params_with_default without _
        cls.complete_params_with_default(params)


def _extend_simul_class_from_path(Simul, path_file):
    """Extend a Simul if needed from a path file (internal API)."""

    path_file = Path(path_file)

    if mpi.rank == 0:
        if path_file.suffix == ".xml":
            # we assume that it is a info_solver.xml file
            info_solver = ParamContainer(path_file=path_file)
            if hasattr(info_solver, "extenders"):
                extenders = info_solver.extenders
            else:
                extenders = []
        else:
            with h5py.File(path_file, "r") as file:
                extenders = list(
                    file["/info_simul/solver"].attrs.get("extenders", [])
                )
    else:
        extenders = None

    if mpi.nb_proc > 1:
        extenders = mpi.comm.bcast(extenders)

    extender_classes = []

    for extender_full_name in extenders:
        module_name, class_name = extender_full_name.rsplit(".", 1)

        try:
            extender_class = import_class(module_name, class_name)
        except ImportError:
            warn(f"ImportError extender class {extender_full_name}.")
        else:
            extender_classes.append(extender_class)

    if extender_classes:
        return extend_simul_class(Simul, extender_classes)
    else:
        return Simul



"""IPython / Jupyter magic commands

.. autoclass:: MagicsCore
   :members:
   :private-members:

"""

from pprint import pprint

from fluiddyn.io.query import query_yes_no
from IPython.core.magic import Magics, line_magic
from IPython.core.magic_arguments import (
    argument,
    magic_arguments,
    parse_argstring,
)

from .loader import available_solvers, import_cls_simul


class MagicsCore(Magics):
    """Magics simplifies the instantiation steps for a Simul object.
    A class variable ``entrypoint_grp`` informs which modules to import from.
    See also :mod:`fluidsim_core.loader`.

    """

    entrypoint_grp = None

    def is_defined(self, varname):
        user_ns = self.shell.user_ns
        return varname in user_ns and user_ns[varname] is not None

    @magic_arguments()
    @argument("solver", nargs="?", default="")
    @argument("-f", "--force-overwrite", action="store_true")
    @line_magic
    def fluidsim(self, line):
        args = parse_argstring(self.fluidsim, line)

        if args.solver == "":
            print("Available solvers:")
            pprint(available_solvers(self.entrypoint_grp), indent=4, compact=True)
            if self.is_defined("params") and self.is_defined("sim"):
                print("\nInitialized:")
                user_ns = self.shell.user_ns
                pprint(
                    dict(
                        params=type(user_ns["params"]), Simul=type(user_ns["sim"])
                    ),
                    indent=4,
                )
            return

        if not args.force_overwrite and (
            self.is_defined("params") or self.is_defined("Simul")
        ):
            if not query_yes_no(
                "At least one of the variables `params` or `Simul` is defined"
                " in your user namespace. Do you really want to overwrite "
                "them?"
            ):
                return

        user_ns = self.shell.user_ns
        Simul = import_cls_simul(args.solver, self.entrypoint_grp)
        params = Simul.create_default_params()
        print(
            "Created Simul class and default parameters for",
            line,
            "-> Simul, params",
        )
        user_ns["Simul"] = Simul
        user_ns["params"] = params

    @magic_arguments()
    @line_magic
    def fluidsim_reset(self, line):
        for varname in ("sim", "params", "Simul"):
            if self.is_defined(varname):
                del self.shell.user_ns[varname]

    @line_magic
    def fluidsim_help(self, line):
        print(4 * " " + self.__doc__)



"""Helper to plot remaining clock time data

Provides:

.. autoclass:: RemainingClockTime
   :members:
   :private-members:
   :undoc-members:

"""

from abc import ABCMeta, abstractmethod
from datetime import timedelta

import numpy as np
import matplotlib.pyplot as plt


class RemainingClockTime(metaclass=ABCMeta):
    @abstractmethod
    def _load_times(self):
        """Load remaining time data.

        - equation_times
        - remaining_clock_times
        - clock_times_per_timestep
        - equation_time_start
        - full_clock_time

        """

    def plot_clock_times(self):
        """Plot the estimated full clock time and clock time per time step."""

        results = self._load_times()
        equation_times = results["equation_times"]

        if len(equation_times) == 0:
            print("No time data in the log file. Can't plot anything.")
            return

        fig, axes = plt.subplots(2, 1, sharex=True)

        remaining_clock_times = results["remaining_clock_times"]

        ax = axes[0]
        ax.plot(equation_times, remaining_clock_times)
        ax.set_ylabel("estimated full clock time (s)")

        if remaining_clock_times[-1] > 0.05 * np.nanmax(remaining_clock_times):
            last_remaining = remaining_clock_times[-1]
            ax.plot(
                equation_times[-1],
                last_remaining,
                "rx",
                label=(
                    f"Last estimation {timedelta(seconds=int(last_remaining))}"
                ),
            )
            ax.legend()

        clock_times_per_timestep = results["clock_times_per_timestep"]
        if clock_times_per_timestep[-1] <= 0.0:
            equation_times = equation_times[:-1]
            clock_times_per_timestep = clock_times_per_timestep[:-1]

        times2 = np.empty(2 * equation_times.size)

        times2[0] = results["equation_time_start"]
        times2[1::2] = equation_times
        times2[2::2] = equation_times[:-1]

        clock_times_per_timestep2 = np.zeros_like(times2)
        clock_times_per_timestep2[::2] = clock_times_per_timestep
        clock_times_per_timestep2[1::2] = clock_times_per_timestep

        ax = axes[1]
        ax.plot(times2, clock_times_per_timestep2)
        ax.set_xlabel("equation time")
        ax.set_ylabel("clock time per time step (s)")
        full_clock_time = results["full_clock_time"]
        ax.set_title(
            f"Full clock time: {timedelta(seconds=int(full_clock_time))}"
        )

        for ax in axes:
            ax.set_ylim(bottom=0)

        fig.suptitle(self.output.summary_simul, fontsize=8)
        fig.tight_layout()

        print(
            "Mean clock time per time step: "
            f"{np.nanmean(clock_times_per_timestep2):.3g} s"
        )
